"""Zellulärer Automat: Conways Game of Life."""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation

# Erzeuge ein leeres quadratisches NxN-Spielfeld. Es soll nur zwei Zustände
# geben:
#          0  - ein leeres Feld
#          1  - ein Feld ist mit einem Lebewesen besetzt
N = 200
feld = np.empty((N, N), dtype=np.int8)

# Definiere die Indizes der Nachbarschaftsbeziehung
nachbarschaft = [(-1, 0), (+1, 0), (0, -1), (0, +1),
                 (-1, -1), (+1, +1), (-1, +1), (+1, -1)]

# Erstelle ein Plot-Fenster und erzeuge die Darstellung des Spielfelds.
fig = plt.figure()
ax = fig.add_subplot(1, 1, 1)
colormap = matplotlib.colors.ListedColormap(["wheat", "red"])
image = ax.imshow(feld, clim=(0, 1), cmap=colormap)
text = fig.text(0.5, 0.99, '',
                horizontalalignment='center',
                verticalalignment='top')
fig.tight_layout()


def init():
    """Initialisiert die Grafikausgabe beipsielsweise nach Größenänderung des
    Ausgabefensters."""
    return image, text


def update(frame):
    """Berechnet den nächsten Zeitschritt und gibt die zu aktualisierenden
    Grafikelemente zurück. """
    global feld, t

    # Aktualisiere das Textfeld
    text.set_text(f'Generation Nr. {t}')

    # Im nullten Zeitschritt soll zunaechst nur der Anfangszustand
    # dargestellt werden.
    if t == 0:
        image.set_array(feld)
        t += 1
        return image, text

    # Erstelle eine Matrix, die die Anzahl der lebenden nächsten Nachbarn
    # einer jeden Zelle enthält. Diese Zahl liegt stets im Bereich von 0 bis 8
    nachbarn = sum(np.roll(feld, (i, j), axis=(0, 1))
                   for (i, j) in nachbarschaft)

    # Wende die Regeln an
    feld[nachbarn > 3] = 0                     # Sterben durch Übervölkerung
    feld[nachbarn < 2] = 0                     # Sterben durch Vereinsamung
    feld[nachbarn == 3] = 1                    # Geburt bei exakt 3 Nachbarn

    # Lebewesen am Rand des Spielfeldes sterben. Man könnte das auch
    # weglassen und hätte damit periodische Randbedingungen implementiert.
    feld[:, 0] = 0           # linker Rand
    feld[:, -1] = 0          # rechter Rand
    feld[0, :] = 0           # oberer Rand
    feld[-1, :] = 0          # unterer Rand

    # Akualisiere das Bild
    image.set_array(feld)

    # Erhöhe den Zähler für die Zeitschritte
    t += 1

    return image, text


# Zeitzähler wird auf Null gesetzt.
t = 0

# Als Anfangsbedingung geben wir ein leeres Spielfeld vor, dass
# mit 5 lebenden Zellen belegt ist, die ein sogenanntes "r-Pentomino" bilden
#
#           be
#          ac
#           d
#
feld[:] = 0
row = N // 2
col = N // 2
feld[row, col] = 1                  # a
feld[row - 1, col + 1] = 1          # b
feld[row, col + 1] = 1              # c
feld[row + 1, col + 1] = 1          # d
feld[row - 1, col + 2] = 1          # e

# Erzeuge die Animation und starte sie.
ani = matplotlib.animation.FuncAnimation(fig, update, init_func=init,
                                         interval=100)
plt.show()