"""Zellulärer Automat für ein einfaches Waldbrandmodell."""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation

# Erzeuge ein quadratisches NxN-Spielfeld.
# Es gibt insgesamt nur 3 mögliche Zustände:
#          0  - ein leeres Feld
#          1  - ein brennender Baum
#          2  - ein intakter Baum
N = 500
feld = np.empty((N, N), dtype=np.int8)

# Wahrscheinlichkeit, dass ein bereits brennender Baum endgültig abbrennt.
p_abbrenn = 1.0

# Wahrscheinlichkeit, dass ein Baum durch einen brennenden Baum in der
# direkten Nachbarschaft Feuer fängt.
p_feuerfang = 1.0

# Wahrscheinlichkeit, dass auf einem leeren Feld ein Baum neu entsteht.
p_wachs = 0.01

# Wahrscheinlichkeit, dass ein Baum von sich aus anfängt zu brennen.
p_spontan = 0.00001

# Definiere die Indizes der Nachbarschaftsbeziehung
nachbarschaft = [(-1, 0), (+1, 0), (0, -1), (0, +1)]

# Erstelle ein Plot-Fenster und erzeuge die Darstellung des Spielfelds.
fig = plt.figure()
ax = fig.add_subplot(1, 1, 1)
colormap = matplotlib.colors.ListedColormap(["black", "red", "green"])
image = ax.imshow(feld, clim=(0, colormap.N - 1), cmap=colormap)
fig.tight_layout()


def init():
    """Initialisiert die Grafikausgabe beipsielsweise nach Größenänderung des
    Ausgabefensters."""
    return image,


def update(frame):
    """Berechnet den nächsten Zeitschritt und gibt die zu aktualisierenden
    Grafikelemente zurück. """
    global feld, t

    # Im nullten Zeitschritt soll zunaechst nur der Anfangszustand
    # dargestellt werden.
    if t == 0:
        image.set_array(feld)
        t += 1
        return image,

    # Erstelle ein Array mit gleichverteilten Zufallszahlen zwischen 0 und 1,
    # das die gleiche Größe hat, wie das Spielfeld
    feld_rand = np.random.random_sample(feld.shape)

    # Erstelle eine Kopie des aktuellen Zustands
    tmp = feld.copy()

    # Brennende Bäume verwandeln sich mit der Wahrscheinlichkeit p_abbrenn
    # in leere Felder.
    feld[(tmp == 1) & (feld_rand < p_abbrenn)] = 0

    # Intakte Bäume entzünden sich mit der Wahrscheinlichkeit p_spontan
    feld[(tmp == 2) & (feld_rand < p_spontan)] = 1

    # Auf leeren Plätzen wachsen mit der Wahrscheinlichkeit p_wachs neue Bäume
    feld[(tmp == 0) & (feld_rand < p_wachs)] = 2

    # Zähle für jeden Eintrag des Feldes die Anzahl der brennenden Bäume
    # in der Nachbarschaft.
    n_feuer = sum(np. roll(tmp == 1, (i, j), axis=(0, 1))
                  for (i, j) in nachbarschaft)

    # Intakte Bäume, in deren Nachbarschaft mindestens ein brennender Baum
    # existiert, fangen mit der Wahrscheinlichkeit p_feuerfang Feuer.
    feld[(tmp == 2) & (n_feuer > 0) & (feld_rand < p_feuerfang)] = 1

    # Am Rand des Feldes können keine Bäume wachsen.
    feld[:, 0] = 0           # linker Rand
    feld[:, -1] = 0          # rechter Rand
    feld[0, :] = 0           # oberer Rand
    feld[-1, :] = 0          # unterer Rand
    
    # Akualisiere das Bild
    image.set_array(feld)

    # Erhöhe den Zähler für die Zeitschritte
    t += 1

    return image,


# Zeitzähler wird auf Null gesetzt.
t = 0

# Wir starten mit einem leeren Feld.
feld[:, :] = 0

# Erzeuge die Animation und starte sie.
ani = matplotlib.animation.FuncAnimation(fig, update, init_func=init,
                                         interval=1, blit=True)
plt.show()
