"""Zellulärer Automat für ein 1-dimensionales Wellenmodell."""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation
import math

# Erzeuge zwei lineare Spielfelder von jweils N Zellen. Das Array z enthält
# die Auslenkung der Zellen aus der Ruhelage und das Array v enthält die
# Schnelle.
N = 800
z = np.empty(N)
v = np.empty(N)

# Der Parameter alpha gibt die Stärke der Kopplung zweier benachbarter Zellen
# an. Man kann zeigen, dass sich die Phasengeschwindigkeit c der Wellen durch
#      c = sqrt(alpha)
# gegeben ist, wobei c in der Einheit "Zellen / Zeitschritt" gemessen wird.
# Für den Wert von alpha = 0.01 ergibt sich also eine Phasengeschwindigkeit
# von 0.1 Zellen pro Zeitschritt.
alpha = 0.01

# Frequenz der Anregung
f = 0.001

# Anzahl der Wellenzüge, die angeregt werden sollen
n_anreg = 6

# Gebe die Wellenlänge aus.
print(f'lambda = {math.sqrt(alpha) / f} Zellen')

# Definiere die Indizes der Nachbarschaftsbeziehung
nachbarschaft = [-1, 1]

# Erstelle ein Plot-Fenster und erzeuge die Darstellung der Welle.
fig = plt.figure()
ax = fig.add_subplot(1, 1, 1)
plot, = ax.plot(z)
ax.set_ylim([-1, 1])
ax.grid()
fig.tight_layout()


def init():
    """Initialisiert die Grafikausgabe beipsielsweise nach Größenänderung des
    Ausgabefensters."""
    return plot,


def update(frame):
    """Berechnet den nächsten Zeitschritt und gibt ein Tupel der zu
    aktualisierenden Grafikelemente zurück."""
    global z, v, t

    # Im nullten Zeitschritt soll zunaechst nur der Anfangszustand
    # dargestellt werden.
    if t == 0:
        plot.set_ydata(z)
        t += 1
        return plot,

    # Die Änderung der Geschwindigkeit ist proportional zur relativen
    # Verschiebung der Nachbarzellen.
    nachbarn = sum(np.roll(z, i) for i in nachbarschaft)
    v += alpha * (nachbarn - z * len(nachbarschaft))

    # Die Änderung der Auslenkung ist proportional zur Geschwindigkeit. Man
    # könnte hier noch einen Proportionalitätsfaktor angeben, das ist aber
    # nicht erforderlich.
    z += v

    # Die erste Zelle wird harmonisch mit einer Frequenz f angeregt. Die
    # Anregung soll nach n Perioden enden.
    if t < n_anreg / f:
        z[0] = 0.25 * math.sin(2 * math.pi * f * t)
    else:
        z[0] = 0

    # Die letzte Zelle stellt ein festes Ende dar.
    z[-1] = 0

    # Alternativ kann man auch ein loses Ende simulieren. Der einfachste Ansatz
    # besteht in der Randbedinung
    # z[-1] = z[-2]

    # Akualisiere das Bild
    plot.set_ydata(z)

    # Gehe zum nächsten Zeitschritt
    t += 1

    return plot,


# Zeitzähler wird auf Null gesetzt.
t = 0

# Anfangsbedingung: Keine Anfangsauslenkung, keine Anfangsgeschwindigkeit
z[:] = 0
v[:] = 0

# Erzeuge die Animation und starte sie.
ani = matplotlib.animation.FuncAnimation(fig, update, init_func=init,
                                         interval=1, blit=True)
plt.show()
