"""Zellulärer Automat für ein einfaches 2D-Wellenmodell.

In diesem konrekten Beispiel wird die Interferenz von zwei kreisförmigen
Wellen dargestellt."""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation
import math

# Erzeuge zwei quadratische NxN-Spielfelder.
# Das Array z enthält die Auslenkung der Zellen aus der Ruhelage und
# das Array v enthält die Schnelle.
N = 700
z = np.empty((N, N))
v = np.empty((N, N))

# Der Parameter alpha gibt die Stärke der Kopplung zweier benachbarter Zellen
# an. Man kann zeigen, dass die Phasengeschwindigkeit c der Wellen durch
#      c = sqrt(alpha)
# gegeben ist, wobei c in der Einheit "Zellen / Zeitschritt" gemessen wird.
# Für den Wert von alpha = 0.04 ergibt sich also eine Phasengeschwindigkeit
# von 0.2 Zellen pro Zeitschritt.
alpha = 0.04

# Frequenz der Anregung in der Einheit "1 / Zeitschritt".
f = 0.008

# Wellenlänge in der Einheit "Zellen".
lamb = math.sqrt(alpha) / f

# Gebe die Wellenlänge aus.
print(f'lambda = {lamb} Zellen')

# Definiere die Indizes der Nachbarschaftsbeziehung
nachbarschaft = [(-1, 0), (+1, 0), (0, -1), (0, +1)]

# Erstelle ein Plot-Fenster und erzeuge die Darstellung des Spielfelds.
fig = plt.figure()
ax = fig.add_subplot(1, 1, 1)
image = ax.imshow(z, clim=(-1, 1), cmap=matplotlib.cm.jet)
fig.tight_layout()


def init():
    """Initialisiert die Grafikausgabe beipsielsweise nach Größenänderung des
    Ausgabefensters."""
    return image,


def update(frame):
    """Berechnet den nächsten Zeitschritt und gibt ein Tupel der zu
    aktualisierenden Grafikelemente zurück."""
    global z, v, t

    # Im nullten Zeitschritt soll zunaechst nur der Anfangszustand
    # dargestellt werden.
    if t == 0:
        image.set_array(z)
        t += 1
        return image,

    # Die Änderung der Geschwindigkeit ist proportional zur relativen
    # Verschiebung der Nachbarzellen.
    nachbarn = sum(np.roll(z, (i, j), axis=(0, 1))
                   for (i, j) in nachbarschaft)
    v += alpha * (nachbarn - z * len(nachbarschaft))
    z += v

    # Dirichlet-Randbedingung
    z[:, 0] = 0
    z[:, -1] = 0
    z[0, :] = 0
    z[-1, :] = 0

    # Anregung durch zwei punktförmige Sender im Abstand 1.5 lambda
    z[int(N/2 - 0.75 * lamb), N//2] = 10 * math.sin(2 * math.pi * f * t)
    z[int(N/2 + 0.75 * lamb), N//2] = 10 * math.sin(2 * math.pi * f * t)

    # Akualisiere das Bild
    image.set_array(z)

    # Gehe zum nächsten Zeitschritt
    t += 1

    return image,


# Zeitzähler wird auf Null gesetzt.
t = 0

# Anfangsbedingung: Keine Anfangsauslenkung, keine Anfangsgeschwindigkeit
z[:, :] = 0
v[:, :] = 0

# Erzeuge die Animation und starte sie.
ani = matplotlib.animation.FuncAnimation(fig, update, init_func=init,
                                         interval=1, blit=True)
plt.show()
