"""Zellulärer Automat für ein einfaches Wärmetransportmodell.

Mit dieser Simulation können wir die Frage beantworten, warum es in Wohnungen
bevorzugt in den Zimmerecken schimmelt."""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation

# Erzeuge ein quadratisches NxN-Spielfeld. Jedes Spielfeld enthält die
# lokale Temperatur.
N = 100
temperatur = np.empty((N, N))

# Der Parameter gibt die Stärke der Diffusion an.
alpha = 0.15           # Temperaturleitfähigkeit
T0 = -10.0             # Temperatur im Außenbereich
T1 = +22.0             # Temperatur im Innenbereich

# Ein 10 Pixel breiter Streifen links und rechts wird auf der Außentemperatur
# gehalten.
index_aussen = np.zeros((N, N), dtype=bool)
index_aussen[:, 0:10] = True
index_aussen[:, 90:] = True

# Zwei rechteckige Felder oben und unten werden auf der Innentemperatur
# gehalten.
index_innen = np.zeros((N, N), dtype=bool)
index_innen[0:30, 30:70] = True
index_innen[70:, 30:70] = True

# Definiere die Indizes der Nachbarschaftsbeziehung
nachbarschaft = [(-1, 0), (+1, 0), (0, -1), (0, +1)]

# Erstelle ein Plot-Fenster und erzeuge die Darstellung des Spielfelds.
fig = plt.figure()
ax = fig.add_subplot(1, 1, 1)
image = ax.imshow(temperatur, clim=(T0, T1), cmap=matplotlib.cm.jet)
ax.set_axis_off()
plt.colorbar(image)
fig.tight_layout()


def init():
    """Initialisiert die Grafikausgabe beipsielsweise nach Größenänderung des
    Ausgabefensters."""
    return image,


def update(frame):
    """Berechnet den nächsten Zeitschritt und gibt ein Tupel der zu
    aktualisierenden Grafikelemente zurück. """
    global temperatur, t

    # Im nullten Zeitschritt soll zunaechst nur der Anfangszustand
    # dargestellt werden.
    if t == 0:
        image.set_array(temperatur)
        t += 1
        return image,

    # Zwischen benachbarten Zellen findet ein Temperaturausgleich statt,
    nachbarn = sum(np.roll(temperatur, (i, j), axis=(0, 1))
                   for (i, j) in nachbarschaft)
    temperatur += alpha * (nachbarn - temperatur * len(nachbarschaft))

    # Berücksichtige die Randbedingungen.
    temperatur[index_aussen] = T0
    temperatur[index_innen] = T1

    # Akualisiere das Bild
    image.set_array(temperatur)

    # Erhöhe den Zähler für die Zeitschritte
    t += 1

    return image,


# Zeitzähler wird auf Null gesetzt.
t = 0

# Im Anfangszustand ist die Temperatur innen T0, die Temperatur außen ist T1
# und die Temperatur in der Wand ist der Mittelwert von T0 und T1.
temperatur[:] = (T0 + T1) / 2
temperatur[index_aussen] = T0
temperatur[index_innen] = T1

# Erzeuge die Animation und starte sie.
ani = matplotlib.animation.FuncAnimation(fig,
                                         update,        
                                         interval=1,  # Zeit pro Bild in ms
                                         blit=True)
plt.show()
