"""Analyse des einfachen Diffusionsmodells.

Es soll die Frage beantwortet werden, ob das Ergebnis des einfachen
Diffusionsmodells auf Basis des zellulären Automaten mit der analytischen
Fundamentallösung der Diffusionsgleichung übereinstimmt."""

import numpy as np
import matplotlib.pyplot as plt

N = 201             # Spielfeldgröße
c0 = 10000.0        # Anfangskonzentration

# Die Konzentration ist am Anfang überall 0. Der mittlere Punkt des Spielfedes
# erhält die vorgegebene Anfangskonzentration
konzentration = np.zeros((N, N))
konzentration[N // 2, N // 2] = c0

# Definiere die Indizes der Nachbarschaftsbeziehung
nachbarschaft = [(-1, 0), (+1, 0), (0, -1), (0, +1)]

# Diffusionskonstante
D = 0.2

# Simuliere eine bestimmte Anzahl von Zeitschritten
for time in range(100):
    nachbarn = sum(np.roll(konzentration, (i, j), axis=(0, 1))
                   for (i, j) in nachbarschaft)
    konzentration += D * (nachbarn - konzentration * len(nachbarschaft))

# Vergleiche die Simulation mit der analytischen Lösung der Diffusionsgleichung.
fig = plt.figure()
ax = fig.add_subplot(1, 1, 1)

# Analytische Lösung der 2-dimensionalen Diffusionsgleichung für eine
# delta-förmige Ausgangssituation
x = np.linspace(-1, 1, N)
y_analyt = c0 / (4 * np.pi * D * time) * \
           np.exp(-(N * x / 2) ** 2 / (4 * D * time))

# Betrachte einen horizontalen Schnitt durch das simulierte Spielfeld.
y_horiz = konzentration[N // 2, :]

# Betrachte einen diagonalen Schnitt durch das simulierte Spielfeld.
# Achtung, im diagonalen Schnitt muss man die Ortsachse mit einem Faktor
# sqrt(2) saklieren.
y_diag = konzentration.diagonal()

# Stelle die drei Konzentrationskurven im Vergleich dar.
ax.plot(x, y_analyt, color='black', linestyle='-', label='Analytisch')
ax.plot(x, y_horiz, color='red', linestyle='--', label='Sim (x)')
ax.plot(x * np.sqrt(2), y_diag,
        color='blue', linestyle='-.', label='Sim (diag)')
ax.set_xlabel('Position / a.u.')
ax.set_ylabel('Konzentration / a.u.')
ax.legend()
plt.show()
