"""Zellulärer Automat für ein einfaches Diffusions-Reaktionsmodell."""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation

# Wir haben nun zwei quadratische Felder. Das Feld "u" enthält die
# Konzentration der ersten Substanz. Das Feld "v" enthält die Konzentration
# der zweiten Substanz. Die Substanzen diffundieren und reagieren miteinander.
N = 300
u = np.empty((N, N))
v = np.empty((N, N))

# Es gibt 4 Parameter
#    Du    : Diffusionskonstante der ersten Substanz.
#    Dv    : Diffusionskonstante der zweiten Substanz.
#    alpha : Reaktionsparameter 1
#    beta  : Reaktionsparameter 2
# Du, Dv, alpha, beta = 0.24, 0.10, 0.036, 0.1            # Fleckenmuster
Du, Dv, alpha, beta = 0.24, 0.10, 0.040, 0.1            # Streifenmuster

# Definiere die Indizes der Nachbarschaftsbeziehung
nachbarschaft = [(-1, 0), (+1, 0), (0, -1), (0, +1)]

# Erstelle ein Plot-Fenster und erzeuge die Darstellung des Spielfelds.
fig = plt.figure()
ax = fig.add_subplot(1, 1, 1)
image = ax.imshow(u, clim=(0, 1), cmap=matplotlib.cm.jet)
ax.set_axis_off()
fig.colorbar(image)
fig.tight_layout()


def init():
    """Initialisiert die Grafikausgabe beipsielsweise nach Größenänderung des
    Ausgabefensters."""
    return image,


def update(frame):
    """Berechnet den nächsten Zeitschritt und gibt ein Tupel der zu
    aktualisierenden Grafikelemente zurück. """
    global u, v, t

    # Im nullten Zeitschritt soll zunaechst nur der Anfangszustand
    # dargestellt werden.
    if t == 0:
        image.set_array(u)
        t += 1
        return image,

    # Da die Grafikausgabe recht langsam ist, berechnen wir immer gleich
    # 15 Zeitschritte, bevor wir das nächste Bild ausgeben.
    dt = 15
    for k in range(dt):
        # Diffusion der ersten Substanz
        nachbarn = sum(np.roll(u, (i, j), axis=(0, 1))
                       for (i, j) in nachbarschaft)
        u += Du * (nachbarn - u * len(nachbarschaft))

        # Diffusion der zweiten Substanz
        nachbarn = sum(np.roll(v, (i, j), axis=(0, 1))
                       for (i, j) in nachbarschaft)
        v += Dv * (nachbarn - v * len(nachbarschaft))

        # In jeder Zelle reagieren u und v miteinander gemäß der folgenden
		# Gleichungen miteinander:
        du = -u * v ** 2 + alpha * (1 - u)
        dv = u * v ** 2 - beta * v
        u += du
        v += dv

    # Akualisiere das Bild
    image.set_data(u)

    # Erhöhe den Zähler für die Zeitschritte
    t += dt

    return image,


# Zeitzähler wird auf Null gesetzt.
t = 0

# Wir starten mit einem linearen Konzentrationsgefällen der beiden
# Substanzen. Dann hat man immer an irgendeiner Stelle des Feldes günstige
# Anfangsbedingungen für die Reaktion.
u, v = np.meshgrid(np.linspace(0, 1, N), np.linspace(0, 1, N))

# Erzeuge die Animation und starte sie.
ani = matplotlib.animation.FuncAnimation(fig, update, init_func=init,
                                         interval=1, blit=True)
plt.show()
