"""Zellulärer Automat für ein einfaches Ising-Modell.

Das Modell beschreibt ein System von Spins, die jeweils nur die Werte +1 oder
-1 annehmen können. Die Spins sind mit ihren nächsten Nachbarn gekoppelt. Von
eiem Zeitschritt zum nächsten verändert ein Spin genau dann seinen Zustand,
wenn dies ohne Änderung der Energie erfolgt (Mikrokanonisches Ensemble). Dazu
müssen in der Nachbarschaft genauso viele Spins den Zustand +1 und den
Zustand -1 haben. """

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation

# Erzeuge ein leeres quadratisches NxN-Spielfeld
N = 500
spin = np.empty((N, N), dtype=np.int8)

# Wir teilen das Spielfeld schachbrettmusterartig in zwei Gruppen ein.
# Damit ist es einfach, die Gesamtenergie festzulegen, indem man einer
# bestimmten Anzahl von Spins aus einer der beiden Gruppen den Wert +1
# zuweist und alle anderen Spins auf -1 setzt.
#
# Das Array gruppe1 ist für alle Spins der ersten Gruppe True und für alle
# Spins der zweiten Gruppe False
gruppe1 = np.zeros(spin.shape, dtype=bool)
gruppe1[1::2, ::2] = True
gruppe1[::2, 1::2] = True

# Definiere die Indices der Nachbarschaftsbeziehung
nachbarschaft = [(-1, 0), (+1, 0), (0, -1), (0, +1)]

# Erstelle ein Plot-Fenster und erzeuge die Darstellung des Spielfelds.
fig = plt.figure()
ax = fig.add_subplot(1, 1, 1)
colormap = matplotlib.colors.ListedColormap(["wheat", "red"])
image = ax.imshow(spin, clim=(-1, 1), cmap=colormap)
text = fig.text(0.5, 0.99, '',
                horizontalalignment='center',
                verticalalignment='top')
ax.set_axis_off()
fig.tight_layout()


def init():
    """Initialisiert die Grafikausgabe beipsielsweise nach Größenänderung des
    Ausgabefensters."""
    return image, text


def update(frame):
    """Berechnet den nächsten Zeitschritt und gibt ein Tupel der zu
    aktualisierenden Grafikelemente zurück."""
    global spin, t

    # Im nullten Zeitschritt soll zunaechst nur der Anfangszustand
    # dargestellt werden. Es darf hier also noch keine Zeitschritt erfolgen.
    if t == 0:
        image.set_array(spin)
        t += 1
        return image, text

    # Ein Spin wird genau dann umgedreht, wenn die Summe der Werte der
    # Nachbarfelder Null ergibt. Damit das funktioniert, ohne die
    # Energieerhaltung zu verletzen, muss man erst alle Spins in der ersten
    # Gruppe behandeln und danach alle Spins in der zweiten Gruppe.

    # Berechne die Summe der Nachbarfelder
    nachbarn = sum(np.roll(spin, (i, j), axis=(0, 1))
                   for (i, j) in nachbarschaft)

    # Drehe alle Spins aus Gruppe 1 um, für die die Summe der Nachbarfelder
    # Null ergibt.
    spin[gruppe1 & (nachbarn == 0)] *= -1

    # Wiederhole den Vorgang für die Spins, die nicht in Gruppe1 sind.
    nachbarn = sum(np.roll(spin, (i, j), axis=(0, 1))
                   for (i, j) in nachbarschaft)
    spin[~gruppe1 & (nachbarn == 0)] *= -1

    # Wir berechnen die Gesamtenergie, um sicherzustellen, dass unser
    # Algorithmus auch wirklich energieerhaltend ist. Man kann damit auch
    # ganz schnell feststellen, dass die Gesamtenergie nicht erhalten ist,
    # wenn man die Teilung in die zwei Gruppen nicht durchführt.
    E = np.mean(-spin[gruppe1] * nachbarn[gruppe1]) / 2

    # Akualisiere das Bild
    image.set_array(spin)
    text.set_text(f't = {t}   E = {E:5.4f}   '
                  f'Mittelwert = {np.mean(spin):5.2f}')

    # Erhöhe den Zähler für die Zeitschritte
    t += 1

    return image, text


# Zeitzähler wird auf Null gesetzt.
t = 0

# Um einen Anfangszustand mit definierter Energie zu erhalten, setzen wir
# zunächst alle Spins auf -1. Wir definieren nun den Anteil f der Spins aus
# der ersten Gruppe , die anfänglich den Zustand +1 haben sollen. Diese Spins
# werden zufällig ausgewählt.
#
# Für f=0 sind alle also Spins mit -1 initialisiert und wir erhalten den
# Zustand mit der mittleren Energie -2 pro Spin. Für f=1 starten wir mit
# einem reinen Schachbrettmuster. In diesem Fall hat man eine mittlere
# Energie +2 pro Spin. Da jeder umgeklappte Spin die gleiche Energie beiträgt
# ergibt sich die Energie pro Spin au dem Anteil f nach der Gleichung:
#
#                E = 4 * (f - 1/2)
#
# Aus theoretischen Überlegungen weiß man, dass es bei E_krit = -sqrt(2)
# zu einem Phasenübergang kommt. Diese kritische Energie entspricht also einem
# Anteil
#
#   f_krit = E_krit / 4 + 1/2
#          = -sqrt(2)/4  + 1/2
#          = 0.146
#
# Für f > f_krit konvergiert das System gegen einen Zustand mit Mittelwert Null
# (keine Gesamtmagnetisierung). Für f < f_krit gibt es eine endliche
# Gesamtmagnetisierung. Im Vortrag wurden die Situationen f = 0.146 + 0.04
# und f = 0.146 - 0.04 dargestellt.
f = 0.146 + 0.04

# Wir setzen zunächst alle Spins auf -1
spin[:, :] = -1

# Wir wählen die Spins, die +1 werden sollen zufällig aus.
rows, cols = np.where(gruppe1)
ind = np.random.permutation(range(rows.size))
ind = ind[0:int(rows.size * f)]
spin[rows[ind], cols[ind]] = +1

# Erzeuge die Animation und starte sie.
ani = matplotlib.animation.FuncAnimation(fig, update, init_func=init,
                                         interval=1)
plt.show()
